# Generated by Django 4.2.3 on 2023-08-07 19:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import permagate.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserPermission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "permission",
                    models.CharField(
                        max_length=255,
                        validators=[permagate.models.validate_permission],
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="permagate_permissions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "permission")},
            },
        ),
        migrations.CreateModel(
            name="GroupPermission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "permission",
                    models.CharField(
                        max_length=255,
                        validators=[permagate.models.validate_permission],
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="permagate_permissions",
                        to="auth.group",
                    ),
                ),
            ],
            options={
                "unique_together": {("group", "permission")},
            },
        ),
    ]



from django.db import models
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, User as UserType
from django.core.exceptions import ValidationError
from .permission import Permission
from typing import Type

User: Type[UserType] = get_user_model()


def validate_permission(permission: str):
    root = Permission.get_root()
    if permission not in root.permission_list:
        raise ValidationError(
            f"{permission} is not a valid permission, check the list of valid permissions"
        )


class UserPermission(models.Model):
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="permagate_permissions"
    )
    permission = models.CharField(validators=[validate_permission], max_length=255)

    class Meta:
        unique_together = ["user", "permission"]


class GroupPermission(models.Model):
    group = models.ForeignKey(
        Group, on_delete=models.CASCADE, related_name="permagate_permissions"
    )
    permission = models.CharField(validators=[validate_permission], max_length=255)

    class Meta:
        unique_together = ["group", "permission"]



from .permission import Permission


root = Permission().register(
    [
        Permission("test", "Test", "Test permission").register(
            [
                Permission("sub1", "Sub perm"),
                Permission("sub2"),
                Permission("sub3").register(
                    [
                        Permission("sub"),
                    ]
                ),
            ]
        ),
        Permission("test2", "Test", "Test perm"),
    ]
)

rootTwo = Permission().register(
    [
        Permission("test", "2nd root", "Test permission").register(
            [
                Permission("sub1", "Sub perm"),
                Permission("sub2"),
            ]
        ),
        Permission("test2", "Test", "Test perm"),
    ]
)



from __future__ import annotations
from django.conf import settings
import importlib
from typing import Optional


class Permission:
    _root: Optional[Permission] = None

    @classmethod
    def get_root(cls, root_path: Optional[str] = None):
        """
        Loads the root permission from the PERMAGATE_PERMISSIONS environmental variable or the specified root path.
        The PERMAGATE_PERMISSIONS variable contains the path to the module that contains the root permission object. An
        optional suffix of the form ":variable_name" may be appended to the path to specify the name of the variable
        referencing the root permission, however if it's not specified, it's assumed the permission object is referenced
        by a variable named "root". Once loaded, the root permission will be cached, unless a custom root path is
        specified.
        :param root_path: A path to the root permission that overrides the value in PERMAGATE_PERMISSIONS
        :return: The root permission object
        """
        if not cls._root or root_path:
            cls._root = cls._load_permission_root(root_path)
        return cls._root

    @staticmethod
    def _load_permission_root(root_path: Optional[str] = None) -> Permission:
        """
        Loads the root permission from the PERMAGATE_PERMISSIONS environmental variable. The PERMAGATE_PERMISSIONS
        variable contains the path to the module that contains the root permission object. An optional suffix of the
        form ":variable_name" may be appended to the path to specify the name of the variable referencing the root
        permission, however if it's not specified, it's assumed the permission object is referenced by a variable named
        "root".
        :param root_path: A path to the root permission that overrides the value in PERMAGATE_PERMISSIONS
        :return: The root permission object
        """
        root_path = root_path or settings.PERMAGATE_PERMISSIONS
        root_variable_name = "root"
        assert isinstance(root_path, str), "PERMAGATE_PERMISSIONS was not initialized!"
        if ":" in root_path:
            components = root_path.split(":")
            assert (
                len(components) == 2
            ), "Invalid PERMAGATE_PERMISSIONS string, it may contain at most one ':'"
            root_path = components[0]
            root_variable_name = components[1]
        module = importlib.import_module(root_path)
        root_permission = getattr(module, root_variable_name)
        assert (
            isinstance(root_permission, Permission) and root_permission.is_root
        ), "PERMAGATE_PERMISSIONS did not point to a root Permissions object"
        return root_permission

    def __init__(
        self,
        key: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """
        Defines a permission. The root permission whose role is to contain all other permissions is signified by a blank
        key.
        :param key: The permission key - should not contain key separator ('.'), inclusive wildcard ('>'), or root
        permission key ('*')
        :param name: The name of the permission, visible in the control interface
        :param description: The permission description, visible in the control interface
        """
        assert (
            not key or "." not in key
        ), "A permission key must be equivalent to one permission string segment"
        assert (
            not key or ">" not in key
        ), "Permission keys cannot contain inclusive wildcards, use those in permission strings assigned to users/groups"
        assert (
            not key or "*" not in key
        ), "The '*' character is reserved. Leave the key blank to create a root permission instance"
        self._key = key
        self.name = name
        self.description = description
        self._parent: Optional[Permission] = None
        self._children: list[Permission] = []

    @property
    def key(self):
        return self._key or "*"

    @property
    def is_root(self):
        """
        A permission instance is the root permission if the key is blank.
        :return: True if this is the root permission
        """
        return not self._key

    @property
    def absolute_permission(self) -> str:
        """
        Determines the full permission string by taking the parent permission into consideration.
        :return: The absolute permission string
        """
        if self._parent and not self._parent.is_root:
            return f"{self._parent.absolute_permission}.{self.key}"
        return self.key

    def register(self, permissions: list[Permission]):
        """
        Registers a list of permissions as child permissions.
        :param permissions: A permission list
        :return:
        """
        for permission in permissions:
            assert permission.key, "The root permission cannot be registered as a child"
            permission._parent = self
        self._children += permissions
        return self

    def exists(self, permission: str):
        """
        Checks if an absolute permission exists.
        :param permission: The permission string, may end with an inclusive wildcard character ('>') or the root
        reference character ('*')
        :return: True if the permission string is defined in this permission tree
        """
        # If it ends with a wildcard, remove the wildcard
        if permission.endswith(">"):
            permission = permission[:-1]

        assert not permission.startswith(
            ">"
        ), "Permission strings cannot start with an inclusive wildcard"

        if permission.startswith("*"):
            assert (
                permission == "*"
            ), "The root permission string may only contain the root permission reference ('*')"

        keys = permission.split(".")
        if len(keys):
            assert len(keys[0]), "Permission string segments cannot be blank"
            key_matches = keys[0] == self.key
            if key_matches or self.is_root:
                if key_matches:
                    keys = keys[1:]

                if len(keys):
                    permission = ".".join(keys)
                    for child in self._children:
                        if child.exists(permission):
                            return True
                elif key_matches:
                    return True  # We're at the last permission segment and the current key matched
        return False

    @property
    def permission_list(self) -> list[str]:
        """
        Retrieves a list of all available permission strings that may be assigned to users. The list includes the
        permission represented by the current instance as well as the child permissions.
        :return:
        """
        permissions = [self.absolute_permission]
        if len(self._children) and not self.is_root:
            permissions.append(f"{self.absolute_permission}>")

        for child in self._children:
            permissions += child.permission_list
        return permissions



from django.test import TestCase
from django.contrib.auth.models import User as UserType, Group
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from .permission import Permission
from .core import has_permission
from .models import UserPermission, GroupPermission, validate_permission
from typing import Type

User: Type[UserType] = get_user_model()

# Create your tests here.


class PermissionLoaderTest(TestCase):
    def test_default_loading(self, path_override=None):
        path_override = path_override or "permagate.test_permissions"
        root = Permission.get_root(path_override)
        self.assertEqual(isinstance(root, Permission), True)
        self.assertEqual(root.is_root, True)

    def test_specific_loading(self):
        self.test_default_loading("permagate.test_permissions:rootTwo")


class PermissionTest(TestCase):
    def setUp(self) -> None:
        self.root = Permission.get_root("permagate.test_permissions")

    def test_root(self):
        self.assertTrue(self.root.exists("test"))
        self.assertTrue(self.root.exists("test.sub1"))
        self.assertTrue(self.root.exists("test.sub2"))
        self.assertFalse(self.root.exists("test.sub4"))
        self.assertTrue(self.root.exists("test2"))
        self.assertTrue(self.root.exists("*"))
        self.assertTrue(self.root.exists("test>"))
        self.assertRaises(AssertionError, self.root.exists, "*.test")
        self.assertRaises(AssertionError, self.root.exists, ">")

    def test_permission_list(self):
        expected_list = [
            "*",
            "test",
            "test>",
            "test.sub1",
            "test.sub2",
            "test.sub3",
            "test.sub3>",
            "test.sub3.sub",
            "test2",
        ]
        self.assertEqual(self.root.permission_list, expected_list)

    def test_model_validator(self):
        self.assertRaises(ValidationError, validate_permission, "t*st")
        self.assertRaises(ValidationError, validate_permission, "test.sub0")
        self.assertIsNone(validate_permission("*"))
        self.assertIsNone(validate_permission("test>"))


class HasPermissionTest(TestCase):
    path = "permagate.test_permissions"

    def setUp(self) -> None:
        Permission.get_root(self.path)
        self.group, _ = Group.objects.get_or_create(name="test")
        self.user, _ = User.objects.get_or_create(username="tester")
        self.userTwo, _ = User.objects.get_or_create(username="tester2")
        self.rootUser, _ = User.objects.get_or_create(username="root_tester")
        self.group.user_set.add(self.user)
        UserPermission.objects.get_or_create(user=self.user, permission="test.sub1")
        GroupPermission.objects.get_or_create(group=self.group, permission="test.sub3")
        UserPermission.objects.get_or_create(user=self.userTwo, permission="test>")
        UserPermission.objects.get_or_create(user=self.rootUser, permission="*")

    def test_user_permissions(self):

        # Perform negative test
        self.assertFalse(has_permission(self.user, "test"))
        self.assertFalse(has_permission(self.user, "test2"))
        self.assertFalse(has_permission(self.user, "test.sub2"))

        # Test direct assignment
        self.assertTrue(has_permission(self.user, "test.sub1"))

        # Test assignment via group
        self.assertTrue(has_permission(self.user, "test.sub3"))

    def test_inclusive_wildcard(self):
        # Accessible via wildcard
        self.assertTrue(has_permission(self.userTwo, "test"))
        self.assertTrue(has_permission(self.userTwo, "test.sub1"))
        self.assertTrue(has_permission(self.userTwo, "test.sub2"))
        self.assertTrue(has_permission(self.userTwo, "test.sub3.sub"))

        # Negative test
        self.assertFalse(has_permission(self.userTwo, "test2"))

    def test_validation(self):
        self.assertRaises(AssertionError, has_permission, self.userTwo, "test>")

    def test_root_permission(self):
        self.assertTrue(has_permission(self.rootUser, "test.sub1"))
        self.assertTrue(has_permission(self.rootUser, "test"))
        self.assertTrue(has_permission(self.rootUser, "test2"))



from django.db.models import QuerySet
from .models import UserPermission, GroupPermission, User
from .permission import Permission
import logging


logger = logging.getLogger(__name__)


def has_permission(user: User, permission: str) -> bool:
    """
    Verifies that a user has a specific permission, or is a member of a group that has the permission. Note that the
    permission string "a.b>" implies that a user has the a.b permission as well as all child permissions.
    :param user: A User model instance
    :param permission: An absolute permission string (does not contain the inclusive wildcard '>')
    :return: True if the user has the specified permission
    """
    root: Permission = Permission.get_root()
    if root.exists(permission):
        user_permission = _has_permission(user.permagate_permissions.all(), permission)
        if not user_permission:
            for group in user.groups.all():
                group_permissions = group.permagate_permissions.all()
                if _has_permission(group_permissions, permission):
                    return True
        else:
            return user_permission
    else:
        logger.warning(f"Checking for permission {permission} that does not exist")
    return False


def _has_permission(
    queryset: QuerySet[UserPermission, GroupPermission],
    permission: str,
    wildcard_only: bool = False,
) -> bool:
    """
    Determines if a given permission is contained in the specified queryset while considering the inclusive wildcard
    operator '>'. The root permission ('*') is also considered.
    :param queryset: The queryset we're operating on
    :param permission: The permission we're looking up
    :param wildcard_only: Only check if the permission string w/ an appended wildcard character is found in the queryset
    :return: True if the queryset contains the permission
    """
    assert (
        ">" not in permission
    ), "A required permission cannot contain an inclusive wildcard '>'"
    assert len(permission) > 0, "The permission string cannot be empty"
    found = (
        not wildcard_only
        and (
            queryset.filter(permission=permission).exists()
            or queryset.filter(permission="*")
        )
    ) or queryset.filter(permission=f"{permission}>").exists()
    if found:
        return True
    path_segments = permission.split(".")
    path_segments.pop()
    # If there are no segments left in the permission string, the permission was not found
    if not path_segments:
        return False
    permission = ".".join(path_segments)
    # If not set, this is our first run, narrow down the queryset (performance gain?)
    if not wildcard_only:
        queryset = queryset.filter(permission__iendswith=">")
    return _has_permission(queryset, permission, True)
